import unittest

import pytest

import gustaf as gus


@pytest.mark.usefixtures("provide_data_to_unittest")
class BasicCallsTest(unittest.TestCase):
    """
    Checks basic calls - calls without specific arguments + without further
    dependencies.
    Does not really
    """

    def test_vertices_basics(self):
        """
        Call available properties/methods-without-args of Vertices.
        """
        v = gus.Vertices(self.V)
        v.vertices = v.vertices
        v.const_vertices
        v.whatami
        # v.unique_vertices()
        v.bounds()
        v.bounds_diagonal()
        v.bounds_diagonal_norm()
        v.select_vertices([[-1, 0.5], [-1, 0.5], [-1, 0.5]])
        v.copy()

        # v.update_vertices()
        # v.remove_vertices()
        # v.merge_vertices()
        # v.showable()
        # v.show()
        # gus.Vertices.concat()

    def test_edges_basics(self):
        """ """
        es = gus.Edges(self.V, self.E)
        es.edges = es.edges
        es.const_edges
        es.whatami
        es.sorted_edges()
        es.unique_edges()
        es.single_edges()
        es.elements = es.elements
        es.const_elements
        es.centers()
        es.referenced_vertices()
        es.dashed()
        es.shrink()
        es.to_vertices()
        es.single_edges()
        es.remove_unreferenced_vertices()

        # es.update_elements()
        # es.update_edges()

    def test_faces_basics(self):
        for fs in (gus.Faces(self.V, self.TF), gus.Faces(self.V, self.QF)):
            fs.edges()
            fs.whatami
            fs.faces = fs.faces
            fs.const_faces
            fs.sorted_faces()
            fs.unique_faces()
            fs.single_faces()

            fs.sorted_edges()
            fs.unique_edges()
            fs.single_edges()
            fs.centers()
            fs.referenced_vertices()
            fs.shrink()
            fs.to_vertices()
            fs.remove_unreferenced_vertices()

            # fs.update_faces()
            # gus.Faces.whatareyou()

    def test_volumes_bascis(self):
        for vs in (gus.Volumes(self.V, self.TV), gus.Volumes(self.V, self.HV)):
            vs.faces()
            vs.whatami
            vs.volumes = vs.volumes
            vs.const_volumes
            vs.sorted_volumes()
            vs.unique_volumes()
            vs.to_faces()

            vs.sorted_edges()
            vs.unique_edges()
            vs.single_edges()
            vs.centers()
            vs.referenced_vertices()
            vs.shrink()
            vs.to_vertices()
            vs.remove_unreferenced_vertices()

            vs.sorted_faces()
            vs.unique_faces()
            vs.single_faces()
            # gus.Faces.whatareyou()



"""gustaf/gustaf/io/ioutils.py.

utils for lord load and his expected expertise, export.
"""
import os


def abs_fname(fname):
    """Checks if fname is abs. If not, returns abs. Tilde safe.

    Parameters
    -----------
    fname: str

    Returns
    --------
    abs_fname: str
      Maybe same as fname, maybe not.
    """
    if os.path.isabs(fname):
        pass
    # elif fname.startswith("~"):
    elif "~" in fname:
        fname = os.path.expanduser(fname)
    else:
        fname = os.path.abspath(fname)

    return fname


def check_and_makedirs(fname):
    """Checks if the directories of the path exists. If not, makedirs!

    Parameters
    -----------
    fname: str

    Returns
    --------
    None
    """
    dirs = os.path.dirname(fname)

    if dirs == "":
        return None

    if not os.path.isdir(dirs):
        os.makedirs(dirs)

    return None



"""gustaf/gustaf/_base.py.

Useful base class for gustaf. Plus some useful decorators.
"""
from gustaf.utils import log


class GustafBase:
    """Base class for gustaf, where logging is nicely wrapped, and some useful
    methods are defined as classmethods..
    """

    __slots__ = ()

    def __init_subclass__(cls, *args, **kwargs):
        """
        Add logger shortcut.
        """
        super().__init_subclass__(*args, **kwargs)
        cls._logi = log.prepended_log("<" + cls.__qualname__ + ">", log.info)
        cls._logd = log.prepended_log("<" + cls.__qualname__ + ">", log.debug)
        cls._logw = log.prepended_log(
            "<" + cls.__qualname__ + ">", log.warning
        )



import numpy as np
import pytest

import gustaf as gus

# frequently used fixtures
all_grids = (
    "vertices",
    "edges",
    "faces_tri",
    "faces_quad",
    "volumes_tet",
    "volumes_hexa",
)


@pytest.fixture
def volumes_hexa333():
    v = [
        [0.0, 0.0, 0.0],
        [0.5, 0.0, 0.0],
        [1.0, 0.0, 0.0],
        [0.0, 0.5, 0.0],
        [0.5, 0.5, 0.0],
        [1.0, 0.5, 0.0],
        [0.0, 1.0, 0.0],
        [0.5, 1.0, 0.0],
        [1.0, 1.0, 0.0],
        [0.0, 0.0, 0.5],
        [0.5, 0.0, 0.5],
        [1.0, 0.0, 0.5],
        [0.0, 0.5, 0.5],
        [0.5, 0.5, 0.5],
        [1.0, 0.5, 0.5],
        [0.0, 1.0, 0.5],
        [0.5, 1.0, 0.5],
        [1.0, 1.0, 0.5],
        [0.0, 0.0, 1.0],
        [0.5, 0.0, 1.0],
        [1.0, 0.0, 1.0],
        [0.0, 0.5, 1.0],
        [0.5, 0.5, 1.0],
        [1.0, 0.5, 1.0],
        [0.0, 1.0, 1.0],
        [0.5, 1.0, 1.0],
        [1.0, 1.0, 1.0],
    ]

    vol = [
        [0, 1, 4, 3, 9, 10, 13, 12],
        [1, 2, 5, 4, 10, 11, 14, 13],
        [3, 4, 7, 6, 12, 13, 16, 15],
        [4, 5, 8, 7, 13, 14, 17, 16],
        [9, 10, 13, 12, 18, 19, 22, 21],
        [10, 11, 14, 13, 19, 20, 23, 22],
        [12, 13, 16, 15, 21, 22, 25, 24],
        [13, 14, 17, 16, 22, 23, 26, 25],
    ]

    return gus.Volumes(v, vol)


update_elements_params = (
    *all_grids[1:-1],  # without vertices and simple hexa
    "volumes_hexa333",
)


@pytest.mark.parametrize("grid", all_grids)
def test_unique_vertices(grid, request):
    """Test unique_vertices. requires scipy."""
    grid = request.getfixturevalue(grid)

    # random vertices
    n_ran = 50
    random_vertices = np.random.random((n_ran, grid.vertices.shape[1]))
    # copy original
    n_original_vertices = len(grid.vertices)
    original_vertices = grid.vertices.copy()
    # assgin new vertices
    #  ==> stacks (original, random, original, random)
    grid.vertices = np.vstack(
        (grid.vertices, random_vertices, original_vertices, random_vertices)
    )
    # check if vertices are correctly set
    assert len(grid.vertices) == int((n_original_vertices + n_ran) * 2)

    unique_vs = grid.unique_vertices(return_intersection=True)

    n_expected_unique = int(n_original_vertices + n_ran)

    # value check

    assert np.allclose(grid.vertices[:n_expected_unique], unique_vs.values)

    # ids check
    assert all(np.arange(n_expected_unique) == unique_vs.ids)

    # inverse check
    assert all(np.tile(np.arange(n_expected_unique), 2) == unique_vs.inverse)

    # intersection check - should include itself as well
    # also, should be sorted, assuming scipy verion is > 1.6
    intersection_list = [*unique_vs.intersection]
    intersection_ref = [
        [i, i + n_expected_unique] for i in range(n_expected_unique)
    ] * 2
    # although these are integers, this is still very nice one-line assert
    assert np.allclose(intersection_list, intersection_ref)


@pytest.mark.parametrize("grid", all_grids)
def test_bounds(grid, request):
    """bounds should give you AABB"""
    grid = request.getfixturevalue(grid)

    # [0, 1]^3 cube
    ref = [[0, 0, 0], [1, 1, 1]]

    # same?
    assert np.allclose(ref, grid.bounds())

    # add some random values of [0, 1)
    n_original_vs = len(grid.vertices)
    n_ran = 50
    random_vertices = np.random.random((n_ran, grid.vertices.shape[1]))
    grid.vertices = np.vstack((grid.vertices, random_vertices))

    # bound shouldn't change
    assert np.allclose(ref, grid.bounds())

    # remove unreferenced if they are elements
    if not grid.kind.startswith("vertex"):
        grid.remove_unreferenced_vertices()
        assert len(grid.vertices) == n_original_vs
        assert np.allclose(ref, grid.bounds())


@pytest.mark.parametrize("grid", all_grids)
def test_update_vertices(grid, request):
    """update vertices should keep only masked values"""
    grid = request.getfixturevalue(grid)

    # make a copy
    test_grid = grid.copy()

    # int based mask - let's keep 3 vertices
    n_original_vs = len(grid.vertices)
    n_vertices_to_keep = 3
    int_mask = np.random.choice(
        np.arange(n_original_vs),
        n_vertices_to_keep,
        replace=False,
    )

    # save ref values for updated vertices
    updated_vs_ref = test_grid.vertices[int_mask]

    # update_vertices
    test_grid.update_vertices(int_mask)

    assert len(test_grid.vertices) == n_vertices_to_keep
    assert np.allclose(updated_vs_ref, test_grid.vertices)

    # make a copy
    test_grid = grid.copy()
    assert len(test_grid.vertices) == n_original_vs

    # bool based mask - same masking as int_mask
    bool_mask = np.zeros(n_original_vs, dtype=bool)
    bool_mask[int_mask] = True

    # bool mask will keep values in sorted way
    reorganize_ref = np.argsort(int_mask)

    # update_vertices
    test_grid.update_vertices(bool_mask)

    assert len(test_grid.vertices) == n_vertices_to_keep
    assert np.allclose(updated_vs_ref[reorganize_ref], test_grid.vertices)

    # check if elements are updated accordingly
    if not grid.kind.startswith("vertex"):
        # a bit more expensive, but should give
        # same result as inverse
        original_elements = grid.elements.copy()
        kept_vertices = original_elements == int_mask[0]
        for v_id in int_mask[1:]:
            kept_vertices |= original_elements == v_id

        assert len(test_grid.elements) == sum(kept_vertices.all(axis=1))


@pytest.mark.parametrize("grid", update_elements_params)
def test_update_elements(grid, request):
    """keep masked elements"""
    grid = request.getfixturevalue(grid)

    n_original_es = len(grid.elements)
    n_elements_to_keep = 3
    int_mask = np.random.choice(
        np.arange(n_original_es),
        n_elements_to_keep,
        replace=False,
    )

    test_grid = grid.copy()
    # save ref values for updated vertices
    updated_es_ref = test_grid.elements[int_mask].copy()

    # update_vertices
    test_grid.update_elements(int_mask)

    assert len(test_grid.elements) == n_elements_to_keep
    assert np.allclose(
        grid.vertices[updated_es_ref], test_grid.vertices[test_grid.elements]
    )

    # make a copy
    test_grid = grid.copy()
    assert len(test_grid.elements) == n_original_es

    # bool based mask - same masking as int_mask
    bool_mask = np.zeros(n_original_es, dtype=bool)
    bool_mask[int_mask] = True

    # bool mask will keep values in sorted way
    reorganize_ref = np.argsort(int_mask)

    # update_vertices
    test_grid.update_elements(bool_mask)

    assert len(test_grid.elements) == n_elements_to_keep
    assert np.allclose(
        grid.vertices[updated_es_ref[reorganize_ref]],
        test_grid.vertices[test_grid.elements],
    )

    # see if vertices are removed.
    # this uses remove_unreferenced_vertices
    leftover_vertex_ids = np.unique(test_grid.elements)
    leftover_vertex_ids_ref = np.unique(grid.elements[int_mask])
    assert len(leftover_vertex_ids) == len(leftover_vertex_ids_ref)

    assert np.allclose(
        test_grid.vertices[leftover_vertex_ids],
        grid.vertices[leftover_vertex_ids_ref],
    )



"""gustaf/create/faces.py
Routines to create faces.
"""

import numpy as np

from gustaf import create, settings, utils
from gustaf.faces import Faces


def box(
    bounds=[[0, 0], [1, 1]], resolutions=[2, 2], simplex=False, backslash=False
):
    """Create structured quadrangle or triangle block mesh.

    Parameters
    -----------
    bounds: (2, 2) array
        Minimum and maximum coordinates.
    resolutions: (2) array
        Vertex count in each dimension.
    simplex: boolean
        If true, Mesh will be triangular (simplex).

    Returns
    --------
    face_mesh: Volumes
    """

    if np.array(bounds).shape != (2, 2):
        raise ValueError("Bounds must have a dimension of (2, 2).")
    if len(resolutions) != 2:
        raise ValueError("Resolutions must have two entries.")
    if not np.greater(resolutions, 1).all():
        raise ValueError("All resolution values must be at least 2.")

    if simplex:
        # create quad mesh as basis
        quad_mesh = box(bounds=bounds, resolutions=resolutions)

        # turn into triangles
        face_mesh = to_simplex(quad_mesh, backslash)

    else:
        vertex_mesh = create.vertices.raster(bounds, resolutions)
        connectivity = utils.connec.make_quad_faces(resolutions)
        face_mesh = Faces(vertex_mesh.vertices, connectivity)

    return face_mesh


def to_simplex(quad, alternate=False):
    """Given quad faces, diagonalize them to turn them into triangles.

    If quad is counterclockwise (CCW), triangle will also be CCW and
    vice versa. Will return a tri-mesh, if input is triangular.
    Default diagonalization looks like this:

    .. code-block::

        (3) *---* (2)
            |  /|
            | / |
            |/  |
        (0) *---* (1)

    resembling 'slash'.

    .. code-block::

        (3) *---* (2)
            |\\  |
            | \\ |
            |  \\|
        (0) *---* (1)

    resembling 'backslash'.

    If you want to alternate the `slash`-direction, set `alternate`-variable.

    Parameters
    ----------
    quad: Faces
      Faces representation which is to be converted from a cubic mesh into a
      simplex mesh.
    alternate: bool
       Alternate between forward and back-slash to avoid "favored" meshing
       direction (important in some analysis problem).

    Returns
    --------
    tri: Faces
        Simplexifyed mesh.
    """

    if not isinstance(quad, Faces):
        raise ValueError(
            "Input to to_simplex needs to be of type Faces, but it's "
            + type(quad)
        )

    if quad.whatami.startswith("quad"):
        # split variants
        split_slash = [[0, 1, 2], [2, 3, 0]]
        split_backslash = [[0, 1, 3], [3, 1, 2]]

        quad_faces = quad.faces
        tf_half = int(quad_faces.shape[0])
        tri_faces = np.full((tf_half * 2, 3), -1, dtype=settings.INT_DTYPE)

        split = split_backslash if alternate else split_slash

        # If backslash assign every other with backslash else only forward
        tri_faces[0:tf_half:2] = quad_faces[0::2, split_slash[0]]
        tri_faces[1:tf_half:2] = quad_faces[1::2, split[0]]
        tri_faces[tf_half::2] = quad_faces[0::2, split_slash[1]]
        tri_faces[(tf_half + 1) :: 2] = quad_faces[1::2, split[1]]

        tri = Faces(
            vertices=quad.vertices.copy(),
            faces=tri_faces,
        )
    else:
        tri = quad
        utils.log.debug(
            "Non quadrilateral mesh provided, return original" " mesh."
        )

    return tri


def to_quad(tri):
    """
        In case current mesh is triangle surface mesh, it splits triangle faces
        into three quad faces by inserting another vertices in the middle of
        each triangle face. Warning: mesh quality could be bad!

        ``(new) Quad Face``
        .. code-block::

            Ref: (node_ind), face_ind

                     (2)
                     / \
                    / 2 \
                (5)/\\   /\\(4)
                  /  \\ /  \
                 / 0  |(6) \
                /_____|___1_\
              (0)   (3)    (1)

            face_ind | node_ind
            ---------|----------
            0        | 0 3 6 5
            1        | 1 4 6 3
            2        | 2 5 6 4

        Parameters
        -----------
        None

        Returns
        --------
        quad_mesh: Mesh
          Only if current mesh is triangle surface mesh. Otherwise, None.
        """
    if tri.elements is None:
        return None

    if not tri.whatami.startswith("tri"):
        return None

    # Form new vertices: 1. mid edge vertices; 2. center vertices.
    edge_mids = tri.to_edges(unique=True).centers()

    # New vertices - current vertices & mid edge vertices & center vertices
    vertices = np.vstack(
        (
            tri.const_vertices,
            edge_mids,
            tri.centers(),
        )
    )
    em_offset = len(tri.vertices)  # edge mid offset
    fc_offset = len(tri.vertices) + len(edge_mids)  # faces center offset

    # New faces - current faces * 3, as each triangle face should produce
    #  3 quad faces
    faces = np.empty((tri.faces.shape[0] * 3, 4), dtype=settings.INT_DTYPE)

    # Assign face ind
    # First col.
    faces[:, 0] = tri.faces.ravel()
    # Second col.
    edge_mid_column = tri.unique_edges().inverse + em_offset
    faces[:, 1] = edge_mid_column
    # Third col.
    faces[:, 2] = np.repeat(np.arange(len(tri.faces)) + fc_offset, 3)
    # Fourth col.
    faces[:, 3] = edge_mid_column.reshape(-1, 3)[:, [2, 0, 1]].ravel()

    return Faces(vertices=vertices, faces=faces)



"""gustaf/gustaf/io/mixd.py.

io functions for mixd.
"""

import os
import struct

import numpy as np

from gustaf.faces import Faces
from gustaf.io.ioutils import abs_fname, check_and_makedirs
from gustaf.utils import log
from gustaf.volumes import Volumes


def load(
    simplex=True, volume=False, fname=None, mxyz=None, mien=None, mrng=None
):
    """mixd load. To avoid reading minf, all the crucial info can be given as
    params. Default input will try to import `mxyz`, `mien`, `mrng` from
    current location and assumes mesh is 2D triangle.

    Parameters
    -----------
    simplex: bool
      Default is True. Is it triangle based?
    volume: bool
      Default is False. Is it 3D?
    fname: str
      Default is None. Specify your mixd file names with ".xns" postfix.
      Ex) "gustaf.xns" will try to load "gustaf.mxyz", "gustaf.mien",
      and "gustaf.mrng"
    mxyz: str
      Default is None.
    mien: str
      Default is None.
    mrng: str
      Default is None. This is optional.

    Returns
    --------
    mesh: Faces or Volumes
    """
    # figure out input type
    specified_input = mxyz is not None  # bare minimum input
    fname_input = (fname is not None) and not specified_input
    default_input = not (fname_input or specified_input)

    if default_input:
        mxyz = "mxyz"
        mien = "mien"
        mrng = "mrng"

    elif fname_input:
        fname = abs_fname(fname)
        fbase, ext = os.path.splitext(fname)

        if os.path.basename(fbase) == "_":
            fbase = fbase[:-1]
        else:
            fbase += "."

        mxyz = fbase + "mxyz"
        mien = fbase + "mien"
        mrng = fbase + "mrng"

    # vertices
    vertices = np.fromfile(mxyz, dtype=">d").astype(np.float64)

    # connec
    connec = None
    try:
        connec = (np.fromfile(mien, dtype=">i") - int(1)).astype(np.int32)
    except BaseException:
        log.debug(f"mien file, `{mien}`, does not exist. Skipping.")

    # boundary conditions
    bcs = dict()
    try:
        bcs_in = np.fromfile(mrng, dtype=">i").astype(np.int32)  # flattened
        uniq_bcs_in = np.unique(bcs_in)
        uniq_bcs_in = uniq_bcs_in[uniq_bcs_in > 0]  # keep only natural nums
        sub_elem_ids = np.arange(bcs_in.size)

        for ubci in uniq_bcs_in:
            bcs.update({str(ubci): sub_elem_ids[bcs_in == ubci]})

    except BaseException:
        log.debug(f"mrng file, `{mrng}`, does not exist. Skipping.")

    # reshape vertices
    vertices = vertices.reshape(-1, 3) if volume else vertices.reshape(-1, 2)

    # reshape connec
    if connec is not None:
        ncol = int(3) if simplex and not volume else int(4)
        ncol = int(8) if ncol == int(4) and volume and not simplex else ncol

        connec = connec.reshape(-1, ncol)

        mesh = Volumes(vertices, connec) if volume else Faces(vertices, connec)

    # bc
    if len(bcs) != 0:
        mesh.BC = bcs

    return mesh


def export(
    mesh,
    fname,
    space_time=False,
):
    """Export in mixd format. Supports triangle, quadrilateral, tetrahedron,
    and hexahedron semi-discrete and (flat) space-time mesh output.

    Parameters
    -----------
    mesh: Faces or Volumes
    fname: str
    space_time : bool
      Export Mesh as Space-Time Slab for discontinuous space-time

    Returns
    --------
    None
    """
    # did you give us an acceptable mesh?
    acceptable_shapes = ["tri", "quad", "tet", "hexa"]
    whatami = mesh.whatami

    if whatami not in acceptable_shapes:
        raise NotImplementedError(
            f"Sorry, we can't export {whatami}-shape in mixd format."
        )

    # prepare export location
    fname = abs_fname(fname)
    check_and_makedirs(fname)

    # basic infos
    dim = mesh.vertices.shape[1]
    big_endian_int = ">i"
    big_endian_double = ">d"

    # prep files
    fbase, ext = os.path.splitext(fname)

    if ext.startswith(".xns"):
        # frequently used case in practice. no base export
        if os.path.basename(fbase) == "_":
            fdir = os.path.dirname(fbase)
            vert_file = os.path.join(fdir, "mxyz")
            connec_file = os.path.join(fdir, "mien")
            bc_file = os.path.join(fdir, "mrng")
            info_file = os.path.join(fdir, "minf")

        else:
            fbase += "."
            vert_file = fbase + "mxyz"
            connec_file = fbase + "mien"
            bc_file = fbase + "mrng"
            info_file = fbase + "minf"

    else:
        raise NotImplementedError("`mixd` format only supports xns.")

    # write v
    with open(vert_file, "wb") as vf:
        for v in mesh.vertices.ravel():
            vf.write(struct.pack(big_endian_double, v))

        if space_time:
            for v in mesh.vertices.ravel():
                vf.write(struct.pack(big_endian_double, v))

    # write connec
    with open(connec_file, "wb") as cf:
        for c in mesh.elements.ravel() + 1:
            cf.write(struct.pack(big_endian_int, c))

    # write bc
    with open(bc_file, "wb") as bf:
        boundaries = make_mrng(mesh)

        for b in boundaries:
            bf.write(struct.pack(big_endian_int, b))

    # write info
    with open(info_file, "w") as infof:  # if and inf... just can't
        infof.write(f"# dim: {dim}\n")
        infof.write(f"# mesh type: {whatami}\n\n")

        st_factor = 2 if space_time else 1
        infof.write(f"nn {int(mesh.vertices.shape[0] * st_factor)}\n")
        infof.write(f"ne {int(mesh.elements.shape[0])}\n")
        infof.write(f"nsd {dim}\n")
        infof.write(f"nen {int(mesh.elements.shape[1] * st_factor)}\n")

        if space_time:
            infof.write("space-time on\n\n\n")
        else:
            infof.write("semi-discrete on\n\n\n")

        # BC guide
        infof.write("# boundary name : referenced number.\n")
        for i, bname in enumerate(mesh.BC.keys()):
            infof.write(f"# {bname} : {i + 1}\n")

        # signature
        infof.write("\n\n\n# MIXD generated using `gustaf`.\n")


def make_mrng(mesh):
    """
    Builds and return mrng array based on `mesh.BC`
    Supports `Faces` and `Volumes`.

    Parameters
    -----------
    mesh: Faces or Volumes
      Number of participating elements

    Returns
    --------
    boundaries : ndarray
      The mrng-array.
    """

    # determine number of sub elements
    whatami = mesh.whatami
    nbelem = 3

    if whatami.startswith("quad") or whatami.startswith("tet"):
        nbelem += 1
    elif whatami.startswith("hexa"):
        nbelem += 3

    # init boundaries with -1, as it is the value for non-boundary.
    # alternatively, they could be (-1 * neighbor_elem_id).
    # But they aren't.
    boundaries = np.empty(mesh.elements.shape[0] * nbelem, dtype=int)
    boundaries[:] = -1

    for i, belem_ids in enumerate(mesh.BC.values()):
        boundaries[belem_ids] = i + 1  # bid starts at 1

    return boundaries



"""gustaf/gustaf/utils/arr.py.

Useful functions for array / point operations. Named `arr`, since
`array` is python library and it sounds funny.
"""

import numpy as np

from gustaf import settings
from gustaf.helpers.raise_if import ModuleImportRaiser

has_funi = has_napf = has_scipy = False
try:
    import funi

    has_funi = True
except ImportError:
    funi = ModuleImportRaiser("funi")
try:
    import napf

    has_napf = True
except ImportError:
    napf = ModuleImportRaiser("napf")
try:
    import scipy

    has_scipy = True
except ImportError:
    scipy = ModuleImportRaiser("scipy")


def make_c_contiguous(array, dtype=None):
    """Make given array like object a c contiguous np.ndarray. dtype is
    optional. If None is given, just returns None.

    Parameters
    -----------
    array: array-like
    dtype: type or str
      (Optional) `numpy` interpretable type or str, describing type.

    Returns
    --------
    c_contiguous_array: np.ndarray
    """
    if array is None:
        return None

    if isinstance(array, np.ndarray):
        if array.flags.c_contiguous:
            if dtype is not None and array.dtype != dtype:
                return array.astype(dtype)

            return array

    if dtype:
        return np.ascontiguousarray(array, dtype=dtype)

    else:
        return np.ascontiguousarray(array)


def enforce_len(value, n_len):
    """Given int, float, np.ndarray, tuple, list, returns an array with n_len
    len(). In case of iterable, it asserts n_len, else, repeats.

    Parameters
    ----------
    value: int, float or iterable
    n_len: int
      Size of desired array

    Returns
    -------
    len_n_array: (n_len,) np.ndarray
    """
    if isinstance(value, (int, float)):
        return np.repeat(value, n_len)
    elif isinstance(value, (np.ndarray, tuple, list)):
        if len(value) != n_len:
            raise ValueError(
                f"Invalid value length ({len(value)}). ",
                f"Expected length is ({n_len})",
            )
        return np.asarray(value)
    else:
        raise TypeError(
            f"Invalid value type ({type(value)}). "
            "Supports {int, float, np.ndarray, tuple, list}."
        )


def unique_rows(
    in_arr,
    return_index=True,
    return_inverse=True,
    return_counts=True,
    dtype_name=None,
):
    """
    Find unique rows using np.unique, but apply tricks. Adapted from
    `skimage.util.unique_rows`.
    url: github.com/scikit-image/scikit-image/blob/main/skimage/util/unique.py/
    Suitable for int types.

    Parameters
    -----------
    in_arr: (n, m) 2D array-like
    return_index: bool
    return_inverse: bool
    return_counts: bool
    dtype_name: str

    Returns
    --------
    unique_arr: (p, q) np.ndarray
    unique_ind: (w,) np.ndarray
    unique_inv: (t,) np.ndarray
    """
    if dtype_name is None:
        dtype_name = settings.INT_DTYPE

    in_arr = make_c_contiguous(in_arr, dtype_name)

    if len(in_arr.shape) != 2:
        raise ValueError("unique_rows can be only applied for 2D arrays")

    in_arr_row_view = in_arr.view(f"|S{in_arr.itemsize * in_arr.shape[1]}")

    unique_stuff = np.unique(
        in_arr_row_view,
        return_index=True,
        return_inverse=return_inverse,
        return_counts=return_counts,
    )
    unique_stuff = list(unique_stuff)  # list, to allow item assignment

    # switch view to original
    unique_stuff[0] = in_arr[unique_stuff[1]]
    if not return_index:
        # pop return index
        unique_stuff.pop(1)

    return unique_stuff


def close_rows(
    arr, tolerance=None, return_intersection=False, nthreads=None, **kwargs
):
    """Similar to unique_rows, but if data type is floats, use this one.
    Performs radius search using KDTree. Currently uses
    `scipy.spatial.cKDTree`.

    Parameters
    -----------
    arr: (n, d) array-like
    tolerance: (float)
      Defaults to None.
    return_intersection: bool
      Default is False. Returns intersection. For vertices with singular
      points, this will take a lot of memory space.
    nthreads: int
      number of concurrent query. In case of napf, concurrent build as well.
      Default is taken from settings.NTHREADS

    Returns
    --------
    unique_arrays: (n, d) np.ndarray
    unique_ids: (m) np.ndarray
    inverse: (n) np.ndarray
    overlapping: list(list)
        id of neighbors within the tolerance.
    """
    if tolerance is None:
        tolerance = settings.TOLERANCE

    if nthreads is None:
        nthreads = settings.NTHREADS

    if has_funi and not return_intersection:
        return (
            *funi.unique_rows(arr, tolerance, True, "l"),
            [],
        )

    if has_napf:
        kdt = napf.KDT(arr, nthread=nthreads)

        # call the function that's prepared for this moment
        return kdt.unique_data_and_inverse(
            tolerance, True, return_intersection, nthread=nthreads
        )

    if has_scipy:
        from scipy.spatial import cKDTree as scipy_KDTree

        # Build kd tree
        kdt = scipy_KDTree(arr)

        # Ball point query, taking tolerance as radius
        neighbors = kdt.query_ball_point(
            arr,
            tolerance,
            return_sorted=True,
        )

        # inverse based on original vertices.
        o_inverse = np.array(
            [n[0] for n in neighbors],
            dtype=settings.INT_DTYPE,
        )

        # unique of o_inverse, and inverse based on that
        (_, uniq_id, inv) = np.unique(
            o_inverse,
            return_index=True,
            return_inverse=True,
        )

        if not return_intersection:
            neighbors = []

        return arr[uniq_id], uniq_id, inv, neighbors

    raise ImportError(
        "gus.utils.arr.close_rows() requires either funi, napf, or "
        "scipy package."
    )


def bounds(arr):
    """Return bounds.

    Parameters
    -----------
    arr: (n, d) array-like

    Returns
    --------
    bounds: (2, d) np.ndarray
    """
    return np.vstack(
        (
            np.min(arr, axis=0).ravel(),
            np.max(arr, axis=0).ravel(),
        )
    )


def bounds_diagonal(arr):
    """Returns diagonal vector of the bounds.

    bounds[1] - bounds[0]

    Parameters
    -----------
    arr: (n, d) array-like

    Returns
    --------
    bounds_diagonal: (n,) np.ndarray
    """
    b = bounds(arr)
    return b[1] - b[0]


def bounds_norm(arr):
    """Returns norm of the bounds.

    Parameters
    -----------
    arr: (n, d) array-like

    Returns
    --------
    bounds_norm: float
    """
    return np.linalg.norm(bounds_diagonal(arr))


def bounds_mean(arr):
    """Returns mean of the bounds.

    Parameters
    -----------
    arr: (n, d) array-like

    Returns
    --------
    bounds_mean: (n,) array-like
    """
    return np.mean(bounds(arr), axis=0)


def select_with_ranges(arr, ranges):
    """Select array with ranges of each column. Always parsed as:

    [[greater_than, less_than], [....], ...]

    Parameters
    -----------
    ranges: (d, 2) array-like
      Takes None.

    Returns
    --------
    ids: (n,) np.ndarray
    """
    masks = []
    for i, r in enumerate(ranges):
        if r is None:
            continue

        else:
            lower = arr[:, i] > r[0]
            upper = arr[:, i] < r[1]
            if r[1] > r[0]:
                masks.append(np.logical_and(lower, upper))
            else:
                masks.append(np.logical_or(lower, upper))

    if len(masks) > 1:
        mask = np.zeros(arr.shape[0], dtype=bool)
        for i, m in enumerate(masks):
            if i == 0:
                mask = np.logical_or(mask, m)
            else:
                mask = np.logical_and(mask, m)

    else:
        mask = masks[0]

    return np.arange(arr.shape[0])[mask]


def rotation_matrix(rotation, degree=True):
    """Compute rotation matrix. Works for both 2D and 3D point sets. In 2D, it
    can rotate along the (virtual) z-axis. In 3D, it can rotate along [x, y,
    z]-axis. Uses `scipy.spatial.transform.Rotation`.

    Parameters
    -----------
    rotation: list or float
      Amount of rotation along [x,y,z] axis. Default is in degrees.
      In 2D, it can be float.
    degree: bool
      (Optional) rotation given in degrees.
      Default is `True`. If `False`, in radian.

    Returns
    --------
    rotation_matrix: np.ndarray (3,3)
    """
    from scipy.spatial.transform import Rotation as R

    rotation = np.asarray(rotation).ravel()

    if degree:
        rotation = np.radians(rotation)

    # 2D
    if len(rotation) == 1:
        return R.from_rotvec([0, 0, rotation]).as_matrix()[:2, :2]

    # 3D
    elif len(rotation) == 3:
        return R.from_rotvec(rotation).as_matrix()


def rotate(arr, rotation, rotation_axis=None, degree=True):
    """Rotates given arrays. Arrays shape[1] should equal to either 2 or 3 For
    more information, see `rotation_matrix()`.

    Parameters
    -----------
    arr: (n, (2 or 3)) list-like
    rotation: list or float
      angle of rotation (around each axis)
    rotation_axis: (n, (2 or 3)) or (2 or 3) list-like
      center of rotation

    Returns
    --------
    rotated_points: (n, d) np.ndarray
    """
    arr = make_c_contiguous(arr, settings.FLOAT_DTYPE)
    if rotation_axis is not None:
        rotation_axis = make_c_contiguous(
            rotation_axis, settings.FLOAT_DTYPE
        ).ravel()

    if rotation_axis is None:
        return np.matmul(arr, rotation_matrix(rotation, degree))

    else:
        rotated_array = arr - rotation_axis
        rotated_array = np.matmul(
            rotated_array, rotation_matrix(rotation, degree)
        )
        rotated_array += rotation_axis

        return rotated_array


def rotation_matrix_around_axis(axis=None, rotation=None, degree=True):
    """Compute rotation matrix given the axis of rotation. Works for both 2D
    and 3D Uses Rodrigues' formula.

    If axis is not specified, 2D rotation matrix is assumed.

    Parameters
    -----------
    axis: list or np.ndarray
      Axis of rotation in 3D
    rotation: float
      angle of rotation in either radiant or degrees
    degree: bool
      (Optional) rotation given in degrees.
      Default is `True`. If `False`, in radian.

    Returns
    --------
    rotation_matrix: np.ndarray (3,3) of np.ndarray (2,2)
    """
    # Assure angle is specified
    if rotation is None:
        raise ValueError("No rotation angle specified.")
    else:
        if degree:
            rotation = np.radians(rotation)

    # Check Problem dimensions
    if axis is None:
        problem_dimension = 2
    else:
        axis = np.asarray(axis).ravel()
        if axis.shape[0] != 3:
            raise ValueError("Axis dimension must be 3D")
        problem_dimension = 3

    # Assemble rotation matrix
    if problem_dimension == 2:
        rotation_matrix = np.array(
            [
                [np.cos(rotation), -np.sin(rotation)],
                [np.sin(rotation), np.cos(rotation)],
            ]
        )
    else:
        # See Rodrigues' formula
        rotation_matrix = np.array(
            [
                [0, -axis[2], axis[1]],
                [axis[2], 0, -axis[0]],
                [-axis[1], axis[0], 0],
            ]
        )
        rotation_matrix = (
            np.eye(3)
            + np.sin(rotation) * rotation_matrix
            + (
                (1 - np.cos(rotation))
                * np.matmul(rotation_matrix, rotation_matrix)
            )
        )

    return rotation_matrix


def is_shape(arr, shape, strict=False):
    """Checks if arr matches given shape. shape can have negative numbers.

    Parameters
    -----------
    arr: np.ndarray
    shape: tuple
    strict: bool
      raises ValueError if shapes do not match

    Returns
    --------
    matches: bool
    """
    arr = np.asanyarray(arr)

    if arr.ndim != len(shape):
        if strict:
            raise ValueError(f"array should be {arr.ndim}D")
        return False

    for i, (a, s) in enumerate(zip(arr.shape, shape)):
        if s < 0:
            continue
        if a != s:
            if strict:
                raise ValueError(f"array should have {s} shape in {i}-D")
            return False

    return True


def is_one_of_shapes(arr, shapes, strict=False):
    """Tuple/list of given shapes, iterates and checks with is_shape. Useful if
    you have multiple acceptable shapes.

    Parameters
    -----------
    arr: np.ndarray
    shapes: tuple or list
      tuple/list of tuple/list
    strict: bool

    Returns
    --------
    matches: bool
    """
    arr = np.asanyarray(arr)
    matches = False
    for s in shapes:
        m = is_shape(arr, s, strict=False)
        if m:
            matches = True

    if not matches:
        if strict:
            raise ValueError(
                f"array's shape {arr.shape} is not one of f{shapes}"
            )
        return False

    return True



from setuptools import setup

from docs.source.handle_markdown import process_file

with open("gustaf/_version.py") as f:
    version = eval(f.read().strip().split("=")[-1])

readme = process_file("README.md", relative_links=False, return_content=True)

setup(
    name="gustaf",
    version=version,
    description="Process and visualize numerical-analysis-geometries.",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="Jaewook Lee",
    author_email="jaewooklee042@gmail.com",
    url="https://github.com/tataratat/gustaf",
    packages=[
        "gustaf",
        "gustaf.utils",
        "gustaf.io",
        "gustaf.create",
        "gustaf.helpers",
    ],
    install_requires=["numpy"],
    extras_require={
        "all": [
            "vedo>=2023.4.3",
            "scipy",
            "meshio",
            "napf>=0.0.5",
            "funi>=0.0.1",
        ]
    },
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Natural Language :: English",
        "Topic :: Scientific/Engineering",
    ],
    license="MIT",
)
